import fetch from 'node-fetch';
import qs from 'qs';

class Endpoint {
    /**
     *
     * @param {Client} client
     */
    constructor(client) {
        this.client = client;
    }
}

class Exception extends Error {
    /**
     *
     * @param {string} message
     * @param {Error|Response|Object} [data]
     */
    constructor(message, data = undefined) {
        super(message);

        this.data = data;
    }
}

class Client {
    static METHOD = {
        GET:    'GET',
        POST:   'POST',
        PATCH:  'PATCH',
        PUT:    'PUT',
        DELETE: 'DELETE',
    };

    /**
     *
     * @param {string} token
     * @param {string} server
     * @param {string} version
     */
    constructor(token, server = 'https://api.loader.io', version = 'v2') {
        this.token   = token;
        this.server  = server;
        this.version = version;
    }

    /**
     *
     * @return {Promise<*>}
     */
    async request(path, method, {parameters = undefined, body = undefined} = {}) {
        const query   = qs.stringify(parameters, {
            encodeValuesOnly: true,
            arrayFormat:      'brackets',
            addQueryPrefix:   true
        });
        const url     = `${this.server}/${this.version}/${path}${query}`;
        const options = {
            method,
            parameters,
            body:    body != null && body instanceof Object ? JSON.stringify(body) : body,
            headers: {'loaderio-auth': this.token},
        };

        if (body != null && body instanceof Object) {
            options.headers['Content-Type'] = 'application/json';
        }

        try {
            /** @type {Response} */
            const response = await fetch(url, options);

            switch (response.status) {
                case 200:
                    return await response.json();

                case 204:
                    return undefined;

                default:
                    const errorMessage = await response.json();
                    //noinspection ExceptionCaughtLocallyJS
                    throw new Exception(`${errorMessage.errors.join('. ')}. - Loader.io request ${url} failed with response code ${response.status} (https://en.wikipedia.org/wiki/List_of_HTTP_status_codes#1xx_informational_response).`, response);
            }
        }
        catch (error) {
            if (error instanceof Exception) {
                throw error;
            }

            throw new Exception(`Loader.io request ${url} failed with error ${error.message}`, error);
        }
    }
}

class Application extends Endpoint {
    static STATUS = {
        VERIFIED:   'verified',
        UNVERIFIED: 'unverified',
    };

    /**
     *
     * @param {Client} client
     * @param {string} app
     * @param {string} app_id
     * @param {string} status
     */
    constructor(client, {app, app_id, status}) {
        super(client);

        this.app_id = app_id;
        this.app    = app;
        this.status = status;
    }

    /**
     *
     * @return {Promise<boolean>}
     */
    async delete() {
        try {
            await this.client.request(`apps/${this.app_id}`, Client.METHOD.DELETE);
        }
        catch {
            return false;
        }

        return true;
    }

    /**
     *
     * @return {Promise<boolean>}
     */
    async verify() {
        if (this.status === Application.STATUS.VERIFIED) {
            return true;
        }

        const data = await this.client.request(`apps/${this.app_id}`, Client.METHOD.POST, {body: `method=http`});

        if (data.message === 'success') {
            this.status = Application.STATUS.VERIFIED;
        }

        return this.status === Application.STATUS.VERIFIED;
    }
}

class Applications extends Endpoint {
    /**
     *
     * @param {string} app this is the domain, for which the test must be done
     * @return {Promise<Application>}
     */
    async create(app) {
        const responseData = await this.client.request('apps', Client.METHOD.POST, {body: `app=${app}`});
        if (responseData === undefined) {
            throw new Exception(`Loader.io app ${app} can not be created.`);
        }

        return new Application(this.client, responseData);
    }

    /**
     *
     * @param {string} id
     * @return {Promise<Application>}
     */
    async get(id) {
        const data = await this.client.request(`apps/${id}`, Client.METHOD.GET);
        if (data === undefined) {
            throw new Exception(`Loader.io app ${id} can not be found.`);
        }

        return new Application(this.client, data);
    }

    /**
     *
     * @return {Promise<Application[]>}
     */
    async list() {
        const data = await this.client.request('apps', Client.METHOD.GET);
        if (data === undefined) {
            return [];
        }

        return data.map((entry) => new Application(this.client, entry));
    }
}

class Authentication {
    static TYPE = {
        HTTP_BASIC_AUTH: 'basic'
    };

    /**
     *
     * @param {string} type
     * @param {string} login
     * @param {string} password
     */
    constructor({type, login, password}) {
        this.type     = type;
        this.login    = login;
        this.password = password;
    }

    /**
     *
     * @return {{password: string, type: string, login: string}}
     */
    toJSON() {
        return {
            type:     this.type,
            login:    this.login,
            password: this.password,
        };
    }
}

class Variable {
    /**
     *
     * @param {string} name
     * @param {string} property
     * @param {string} source
     */
    constructor({name, property, source}) {
        this.name     = name;
        this.property = property;
        this.source   = source;
    }

    /**
     *
     * @return {{name: string, property: string, source: string}}
     */
    toJSON() {
        return {
            name:     this.name,
            property: this.property,
            source:   this.source,
        };
    }
}

class Url {
    static TYPE = {
        GET:    'GET',
        POST:   'POST',
        PUT:    'PUT',
        PATCH:  'PATCH',
        DELETE: 'DELETE',
    }

    /**
     *
     * @param {string} url
     * @param {string} [truncated_url]
     * @param {string|null} [raw_post_body = null]
     * @param {string} [request_type]
     * @param {string|null} [payload_file_url = null]
     * @param {Object.<string, string>} [headers = null]
     * @param {Object.<string, string>} [request_params = null]
     * @param {Variable[]} [variables = null]
     * @param {Object|Authentication|null} [authentication]
     */
    constructor({
                    url,
                    truncated_url = null,
                    raw_post_body = null,
                    request_type = Url.TYPE.GET,
                    payload_file_url = null,
                    headers = null,
                    request_params = null,
                    variables = null,
                    authentication = null,
                }) {
        this.url              = url;
        this.truncated_url    = truncated_url;
        this.raw_post_body    = raw_post_body;
        this.request_type     = request_type;
        this.payload_file_url = payload_file_url;
        this.headers          = headers;
        this.request_params   = request_params;
        this.variables        = variables == null ? variables : variables.map((variable) => variable instanceof Variable ? variable : new Variable(variable));
        this.authentication   = authentication != null && authentication instanceof Object && (authentication instanceof Authentication) === false ? new Authentication(authentication) : authentication;
    }

    /**
     *
     * @return {{headers: Object<string, string>, variables: *[], request_type: string, raw_post_body: (string|null), request_params: Object<string, string>, truncated_url: string, payload_file_url: (string|null), url: string, authentication: {password: string, type: string, login: string}}}
     */
    toJSON() {
        return {
            url:              this.url,
            truncated_url:    this.truncated_url || undefined,
            raw_post_body:    this.raw_post_body || undefined,
            request_type:     this.request_type,
            payload_file_url: this.payload_file_url || undefined,
            headers:          this.headers || undefined,
            request_params:   this.request_params || undefined,
            variables:        this.variables instanceof Array ? this.variables.map((variable) => variable.toJSON()) : undefined,
            authentication:   this.authentication instanceof Authentication ? this.authentication.toJSON() : undefined,
        };
    }
}

class Result {
    static STATUS = {
        READY:     'ready',
        NOT_READY: 'not_ready',
    };

    /**
     *
     * @param {string} result_id
     * @param {string|Date} started_at
     * @param {string} status
     * @param {string} public_results_url
     * @param {number} success
     * @param {number} error
     * @param {number} timeout_error
     * @param {number} network_error
     * @param {number} data_sent
     * @param {number} data_received
     * @param {number} avg_response_time
     * @param {number} avg_error_rate
     */
    constructor({
                    result_id,
                    started_at,
                    status,
                    public_results_url,
                    success,
                    error,
                    timeout_error,
                    network_error,
                    data_sent,
                    data_received,
                    avg_response_time,
                    avg_error_rate
                }) {
        this.result_id          = result_id;
        this.started_at         = started_at instanceof Date ? started_at : new Date(started_at);
        this.status             = status;
        this.public_results_url = public_results_url;
        this.success            = success;
        this.error              = error;
        this.timeout_error      = timeout_error;
        this.network_error      = network_error;
        this.data_sent          = data_sent;
        this.data_received      = data_received;
        this.avg_response_time  = avg_response_time;
        this.avg_error_rate     = avg_error_rate;
    }
}

class Results extends Endpoint {
    /**
     *
     * @param {Client} client
     * @param {string} testId
     */
    constructor(client, testId) {
        super(client);

        this.testId = testId;
    }

    /**
     *
     * @param {string} id
     * @return {Promise<Result>}
     */
    async get(id) {
        const data = await this.client.request(`tests/${this.testId}/results/${id}`, Client.METHOD.GET);
        if (data === undefined) {
            throw new Exception(`Loader.io result ${id} can not be found.`);
        }

        return new Result(data);
    }

    /**
     * @return {Promise<Result[]>}
     */
    async list() {
        const data = await this.client.request(`tests/${this.testId}/results`, Client.METHOD.GET);
        if (data === undefined) {
            return [];
        }

        return data.map((entry) => new Result(entry));
    }
}

class Test extends Endpoint {
    static STATUS = {
        PENDING:  'pending',
        RUNNING:  'running',
        COMPLETE: 'complete',
    };

    static TYPE = {
        CLIENTS_PER_TEST:     'per-test',
        CLIENTS_PER_SECOND:   'per-second',
        MAINTAIN_CLIENT_LOAD: 'maintain-load',
    };

    /**
     * @var {Results}
     */
    #results;

    /**
     *
     * @return {Results}
     */
    get results() {
        if (this.#results === undefined) {
            this.#results = new Results(this.client, this.test_id);
        }

        return this.#results;
    }

    /**
     *
     * @param {Client} client
     * @param {string} test_id
     * @param {string} name
     * @param {string} domain
     * @param {string} status
     * @param {string} test_type
     * @param {Url[]} urls
     * @param {number} duration
     * @param {number} initial
     * @param {number} total
     * @param {number} timeout
     * @param errorThreshold
     * @param {string} callback
     * @param {string} callback_email
     * @param {string|Date} scheduled_at
     * @param {string} notes
     * @param {string[]} tag_names
     */
    constructor(client,
                {
                    name,
                    duration,
                    timeout,
                    notes,
                    tag_names,
                    initial,
                    total,
                    status,
                    test_id,
                    test_type,
                    callback,
                    callback_email,
                    scheduled_at,
                    domain,
                    urls,
                }
    ) {
        super(client);

        // API delivers different type values as used for creation
        switch (test_type) {
            case 'Clients per second':
            case Test.TYPE.CLIENTS_PER_SECOND:
                test_type = Test.TYPE.CLIENTS_PER_SECOND;
                break;

            case 'Cycling':
            case Test.TYPE.MAINTAIN_CLIENT_LOAD:
                test_type = Test.TYPE.MAINTAIN_CLIENT_LOAD;
                break;

            case 'Non-Cycling':
            case Test.TYPE.CLIENTS_PER_TEST:
            default:
                test_type = Test.TYPE.CLIENTS_PER_TEST;
                break;
        }

        this.name           = name;
        this.duration       = duration;
        this.timeout        = timeout;
        this.notes          = notes;
        this.tag_names      = tag_names; // missing in get and list of API
        this.initial        = initial;
        this.total          = total;
        this.status         = status;
        this.test_id        = test_id;
        this.test_type      = test_type;
        this.callback       = callback;
        this.callback_email = callback_email;
        this.scheduled_at   = scheduled_at instanceof Date ? scheduled_at : new Date(scheduled_at);
        this.domain         = domain;
        this.urls           = (urls || []).map((url) => url instanceof Url ? url : new Url(url));
    }

    /**
     *
     * @return {Promise<boolean>}
     */
    async run() {
        if (this.status !== Test.STATUS.RUNNING) {
            const responseData = await this.client.request(`tests/${this.test_id}/run`, Client.METHOD.PUT);
            this.status        = responseData.status;
        }

        return this.status === Test.STATUS.RUNNING;
    }

    /**
     *
     * @return {Promise<boolean>}
     */
    async stop() {
        if (this.status === Test.STATUS.RUNNING) {
            const responseData = await this.client.request(`tests/${this.test_id}/stop`, Client.METHOD.PUT);
            this.status        = responseData.status;
        }

        return this.status !== Test.STATUS.RUNNING;
    }

    /**
     *
     * @return {{notes: string, initial: number, scheduled_at: string, timeout: number, test_type: string, duration: number, total: number, callback_email: string, urls: *[], domain: string, name: string, callback: string, status: string, test_id: string}}
     */
    toJSON() {
        return {
            name:           this.name,
            duration:       this.duration,
            timeout:        this.timeout,
            notes:          this.notes,
            initial:        this.initial,
            total:          this.total,
            status:         this.status,
            test_id:        this.test_id,
            test_type:      this.test_type,
            callback:       this.callback,
            callback_email: this.callback_email,
            scheduled_at:   this.scheduled_at.toJSON(),
            domain:         this.domain,
            urls:           this.urls.map((url) => url.toJSON()),
        };
    }
}

/**
 * @typedef {Object} LoaderIOTestCreateData
 * @property {string} name
 * @property {string} [test_type = Test.TYPE.CYCLING]
 * @property {Url[]} urls
 * @property {number} duration
 * @property {number} [initial = 1] will be ignored for non-cycling tests
 * @property {number} [total = 25]
 * @property {number} [timeout = 10000]
 * @property {number} [error_threshold = 50]
 * @property {string} [callback]
 * @property {string} [callback_email]
 * @property {Date} [scheduled_at]
 * @property {string} [notes]
 * @property {string[]} [tage_names]
 */

class Tests extends Endpoint {
    /**
     * @param {LoaderIOTestCreateData} data
     * @return {Promise<Test>}
     */
    async create(data) {
        const responseData = await this.client.request('tests', Client.METHOD.POST, {
            body: {
                test_type:       Test.TYPE.CLIENTS_PER_SECOND,
                initial:         1,
                total:           25,
                timeout:         10000,
                error_threshold: 50,
                ...data,
                urls:            data.urls.map((url) => url instanceof Url ? url.toJSON() : url)
            }
        });

        if (responseData?.message !== 'success') {
            throw new Exception(`Loader.io test ${data.name} can not be created.`);
        }

        return this.get(responseData.test_id);
    }

    /**
     *
     * @param {string} id
     * @return {Promise<Test>}
     */
    async get(id) {
        const data = await this.client.request(`tests/${id}`, Client.METHOD.GET);
        if (data === undefined) {
            throw new Exception(`Loader.io test ${id} can not be found.`);
        }

        return new Test(this.client, data);
    }

    /**
     *
     * @return {Promise<Test[]>}
     */
    async list() {
        const data = await this.client.request('tests', Client.METHOD.GET);
        if (data === undefined) {
            return [];
        }

        return data.map((entry) => new Test(this.client, entry));
    }
}

class Servers extends Endpoint {
    /**
     *
     * @return {Promise<string[]>}
     */
    async list() {
        const data = await this.client.request('servers', Client.METHOD.GET);

        return data?.ip_addresses || [];
    }
}

class LoaderIO {
    /**
     *
     * @param {string} token
     * @param {string} server
     * @param {string} version
     */
    constructor({token, server = 'https://api.loader.io', version = 'v2'}) {
        const client      = new Client(token, server, version);

        this.applications = new Applications(client);
        this.tests        = new Tests(client);
        this.servers      = new Servers(client);
    }
}

export default LoaderIO;
export { Application, Applications, Authentication, Exception, Result, Results, Servers, Test, Tests, Url, Variable };
